import rich_click as click

from click import Context


class LazyGroup(click.Group):

    def list_commands(self, ctx):
        return ["build", "init", "validate", "publish", "stage", "wizard"]

    def get_command(self, ctx, cmd_name):
        from .build import build
        from .publish import publish
        from .stage import stage
        from .init import init
        from .validate import validate

        if cmd_name == "build":
            return build
        elif cmd_name == "init":
            return init
        elif cmd_name == "validate":
            return validate
        elif cmd_name == "publish":
            return publish
        elif cmd_name == "stage":
            return stage
        return None


@click.group(cls=LazyGroup)
@click.pass_context
def kabinet(ctx: Context) -> None:
    """Deploy the arkitekt_next app with Port

    The port deployer is an arkitekt_next plugin service, which allows you to deploy your arkitekt_next app to
    any arkitekt_next instance and make it instantly available to the world. Port uses docker to containerize
    your application and will publish it locally to your dockerhub account, and mark it locally as
    deployed. People can then use your github repository to deploy your app to their arkitekt_next instance.

    """

    pass



"""
This module contains the types for the apps
depending on the builder used.
This module imports all the apps and their types
and sets them as attributes on the App class, if they are available.
If they are not available, they are set to Any, so that we can add
an import exception to the app.


"""

import logging
from typing import Any, Dict, TYPE_CHECKING
from koil import unkoil
from koil.composition import Composition
from fakts_next import Fakts


if TYPE_CHECKING:
    from rekuest_next.rekuest import RekuestNext


logger = logging.getLogger(__name__)


class App(Composition):
    """An app that is built with the easy builder"""

    fakts: Fakts
    services: Dict[str, Any]

    @property
    def rekuest(self) -> "RekuestNext":
        """Get the rekuest service"""
        if "rekuest" not in self.services:
            raise ValueError("Rekuest service is not available")
        return self.services["rekuest"]

    def run(self):
        return unkoil(self.rekuest.arun)

    async def arun(self):
        return await self.rekuest.arun()

    def run_detached(self):
        """Run the app detached"""
        return self.rekuest.run_detached()

    def register(self, *args, **kwargs):
        """Register a service"""

        self.rekuest.register(*args, **kwargs)

    def register_startup(self, *args, **kwargs):
        """Register a startup service"""
        self.rekuest.register_startup(*args, **kwargs)

    def register_background(self, *args, **kwargs):
        """Register a background service"""
        self.rekuest.register_background(*args, **kwargs)

    async def __aenter__(self):
        await super().__aenter__()
        for service in self.services.values():
            await service.__aenter__()

        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        for service in self.services.values():
            await service.__aexit__(exc_type, exc_value, traceback)



from functools import partial
from importlib import import_module, reload
import asyncio

from click import Context
from arkitekt_next import App
from watchfiles import awatch, Change
from rich.panel import Panel
from rich.console import Console
from watchfiles.filters import PythonFilter
import os
import sys
import inspect
from rekuest_next.definition.registry import get_default_definition_registry
from rekuest_next.agents.hooks.registry import get_default_hook_registry
from typing import MutableSet, Tuple, Any, Set
from arkitekt_next.cli.ui import construct_changes_group, construct_app_group
from arkitekt_next.cli.commands.run.utils import import_builder
from arkitekt_next.cli.types import Manifest
from arkitekt_next.app.app import App
import rich_click as click
from arkitekt_next.cli.options import (
    with_fakts_next_url,
    with_builder,
    with_token,
    with_instance_id,
    with_headless,
    with_log_level,
    with_redeem_token,
    with_skip_cache,
)
from arkitekt_next.cli.vars import get_console, get_manifest


class EntrypointFilter(PythonFilter):
    """Checks if the entrypoint is changed"""

    def __init__(self, entrypoint: str, *args, **kwargs) -> None:
        """A filter that checks if the entrypoint is changed

        Parameters
        ----------
        entrypoint : str
            The entrypoint to check
        """
        super().__init__(*args, **kwargs)
        self.entrypoint = entrypoint

    def __call__(self, change: Change, path: str) -> bool:
        """Checks if any of the python filters are changed
        _description_
                Parameters
                ----------
                change : Change
                    The change type
                path : str
                    The causing path

                Returns
                -------
                bool
                    Should we reload?
        """
        x = super().__call__(change, path)
        if not x:
            return False

        basename = os.path.basename(path)
        module_name = basename.split(".")[0]

        return module_name == self.entrypoint


class DeepFilter(PythonFilter):
    """Checks if any of the python filters are changed"""

    def __call__(self, change: Change, path: str) -> bool:
        """Checks if any of the python filters are changed

        Parameters
        ----------
        change : Change
            The change type
        path : str
            The causing path

        Returns
        -------
        bool
            Should we reload?
        """
        return super().__call__(change, path)


async def run_app(app: App) -> None:
    """A helper function to run the app

    Parameters
    ----------
    app : App
        The app to run

    """

    rekuest = app.services.get("rekuest")
    if not rekuest:
        raise Exception("No rekuest service found. We need this to run the app.")

    async with app:
        await rekuest.arun()


def reload_modules(reloadable_modules) -> None:
    """Reloads the modules in the reloadable_modules set"""
    for module in reloadable_modules:
        reload(sys.modules[module])


def check_deeps(changes: Set[Tuple[Change, str]]) -> Set[str]:
    """Checks if any of the changes
    are happening in a module that is installed
    and returns the modules that should be reloaded



    Parameters
    ----------
    changes : Set[ Tuple[Change, str] ]
        The changes to check

    Returns
    -------
    Set[str]
        A set of modules that should be reloaded
    """
    normalized = [os.path.normpath(file) for modified, file in changes]

    reloadable_modules = set()

    for key, v in sys.modules.items():
        try:
            filepath = inspect.getfile(v)
        except OSError:
            continue
        except TypeError:
            continue

        for i in normalized:
            if filepath.startswith(i):
                reloadable_modules.add(key)

    return reloadable_modules


def reset_structure() -> None:
    """Resets the default defintiion rgistry and all
    regitered nodes"""
    get_default_definition_registry().actor_builders.clear()
    get_default_hook_registry().reset()


def is_entrypoint_change(
    changes: MutableSet[Tuple[Any, str]], entrypoint_real_path: str
) -> bool:
    for change, path in changes:
        if os.path.normpath(path) == entrypoint_real_path:
            return True
    return False


def callback(console: Console, future: asyncio.Task[None]):
    if future.cancelled():
        return
    else:
        has_exception = future.exception()

        if not has_exception:
            panel = Panel(
                "App finished running", style="bold yellow", border_style="yellow"
            )
            console.print(panel)
        else:
            try:
                raise has_exception
            except Exception:
                console.print_exception()
                panel = Panel("Error running App", style="bold red", border_style="red")
                console.print(panel)


async def run_dev(
    console: Console,
    manifest: Manifest,
    version: str | None = None,
    builder: str = "arkitekt_next.builders.easy",
    deep: bool = False,
    **builder_kwargs,
):
    entrypoint = manifest.entrypoint
    identifier = manifest.identifier
    version = version or "dev"
    entrypoint_file = f"{manifest.entrypoint}.py"
    os.path.realpath(entrypoint_file)

    builder_func = import_builder(builder)

    generation_message = "[not bold white]This is a development tool for arkitekt_next apps. It will watch your app for changes and reload it when it detects a change. It will also print out the current state of your app.[/]"

    if deep:
        generation_message += "\n\n - [not bold white][b]Deep mode[/] is enabled. This will watch all your installed packages for changes and reload them if they are changed.[/]"
    else:
        generation_message += "\n\n - [not bold white][b]Deep mode[/] is disabled. This will only watch your entrypoint for changes.[/]"

    panel = Panel(
        generation_message,
        style="bold green",
        border_style="green",
        title="ArkitektNext Dev Mode",
    )
    console.print(panel)

    try:
        module = import_module(manifest.entrypoint)

    except Exception:
        console.print_exception()
        panel = Panel(
            f"Error while importing your entrypoint please fix your file {entrypoint} and save",
            style="bold red",
            border_style="red",
        )
        console.print(panel)
        module = None

    current_run: asyncio.Future[None] | None = None
    # This is the main task that is running the app

    try:
        app: App = builder_func(
            identifier=identifier,
            version=version,
            logo=manifest.logo,
            **builder_kwargs,
        )
        group = construct_app_group(app)
        panel = Panel(group, style="bold green", border_style="green")
        console.print(panel)

        current_run = asyncio.create_task(run_app(app))
        current_run.add_done_callback(partial(callback, console))
    except Exception:
        console.print_exception()
        panel = Panel(
            "Error building initial App", style="bold red", border_style="red"
        )
        console.print(panel)

    async for changes in awatch(
        ".",
        watch_filter=EntrypointFilter(entrypoint) if not deep else DeepFilter(),
        debounce=2000,
        step=500,
    ):
        if deep:
            #
            to_be_reloaded = check_deeps(changes)
            if not to_be_reloaded:
                continue
        else:
            to_be_reloaded: Set[str] = set()

        group = construct_changes_group(changes)
        panel = Panel(group, style="bold blue", border_style="blue")
        console.print(panel)

        console.print(panel)
        # Cancelling the app
        if not current_run or current_run.done():
            pass

        else:
            current_run.cancel()
            panel = Panel(
                "Cancelling latest version", style="bold yellow", border_style="yellow"
            )
            console.print(panel)
            try:
                await current_run

            except asyncio.CancelledError:
                pass

        # Restarting the app
        try:
            with console.status("Reloading module..."):
                reset_structure()

                if not module:
                    module = import_module(entrypoint)
                else:
                    if deep:
                        reload_modules(to_be_reloaded)
                    else:
                        reload(module)
        except Exception:
            console.print_exception()
            panel = Panel(
                "Reload unsucessfull please fix your app and save",
                style="bold red",
                border_style="red",
            )
            console.print(panel)
            continue

        try:
            app = builder_func(
                identifier=identifier,
                version=version,
                logo=manifest.logo,
                **builder_kwargs,
            )
            group = construct_app_group(app)
            panel = Panel(group, style="bold green", border_style="green")
            console.print(panel)

            current_run = asyncio.create_task(run_app(app))
            current_run.add_done_callback(partial(callback, console))
        except Exception:
            console.print_exception()
            panel = Panel(
                "Error building reloaded App", style="bold red", border_style="red"
            )
            console.print(panel)


@click.command()
@with_fakts_next_url
@with_builder
@with_token
@with_instance_id
@with_redeem_token
@with_headless
@with_log_level
@with_skip_cache
@click.option(
    "--deep",
    help="Should we check the whole directory for changes and reload them when changes?",
    is_flag=True,
)
@click.pass_context
def dev(ctx: Context, **kwargs):
    """Runs the app in dev mode (with hot reloading)

    Running the app in dev mode will automatically reload the app when changes are detected.
    This is useful for development and debugging.
    """

    manifest = get_manifest(ctx)
    console = get_console(ctx)

    asyncio.run(run_dev(console, manifest, **kwargs))



"""
This module contains the types for the apps
depending on the builder used.
This module imports all the apps and their types
and sets them as attributes on the App class, if they are available.
If they are not available, they are set to Any, so that we can add
an import exception to the app.


"""

from typing import Any, Protocol
import logging
from typing import Dict, TYPE_CHECKING
from koil.composition import KoiledModel
from fakts_next import Fakts


if TYPE_CHECKING:
    from rekuest_next.rekuest import RekuestNext


logger = logging.getLogger(__name__)


class App(Protocol):
    """An app that is built with the easy builder"""

    fakts: Fakts
    services: Dict[str, KoiledModel]

    @property
    def rekuest(self) -> "RekuestNext":
        """Get the rekuest service"""
        ...

    def run(self): ...

    async def arun(self): ...

    def run_detached(self):
        """Run the app detached"""
        return self.rekuest.run_detached()

    def register(self, *args, **kwargs):
        """Register a service"""
        ...

    async def __aenter__(self): ...

    async def __aexit__(self, exc_type, exc_value, traceback): ...


class Builder(Protocol):
    """A protocol for a builder class.

    This protocol defines the methods that a builder class must implement.
    """

    def __call__(self, *args: Any, **kwds: Any) -> Any: ...



""" An example of a simple template for Arkitekt Next"""

from arkitekt_next import register
import time
from typing import Generator


@register
def generate_n_string(n: int = 10, timeout: int = 2) -> Generator[str, None, None]:
    """Generate N Strings

    This function generates {{n}} strings with a {{timeout}} ms timeout between each string


    Parameters
    ----------
    n : int, optional
        The number of iterations, by default 10
    timeout : int, optional
        The timeout, by default 2

    Returns
    -------
    str
        A string with Hello {n}
    """
    for i in range(n):
        print(i)
        time.sleep(timeout)
        yield f"Hello {i}"


@register
def append_world(hello: str) -> str:
    """Append World

    This function appends world to the input string

    Parameters
    ----------
    hello : str
        The input string

    Returns
    -------
    str
        {{hello}} World
    """ """"""
    return hello + " World"


@register
def print_string(input: str) -> str:
    """Print String

    This function prints the input string to
    the console

    Parameters
    ----------
    input : str
        The input string

     Returns
    -------
    str
        The printed string
    """
    print(input)
    return input



from importlib import import_module
from typing import Callable
from arkitekt_next.app.app import App


def import_builder(builder: str) -> Callable[..., App]:
    """Import a builder function from a module.

    Parameters
    ----------
    builder : str
        The builder function to import, in the format "module.function".

    Returns
    -------
    Callable[..., App]
        The imported builder function.

    """

    module_path, function_name = builder.rsplit(".", 1)
    module = import_module(module_path)
    function = getattr(module, function_name)
    return function



from arkitekt_next import register
from mikro_next.api.schema import Image, from_array_like, PartialDerivedViewInput


@register
def max_intensity_projection(image: Image) -> Image:
    """Z-Project the Maximum Intensity

    This function projects the maximum intensity of the input image
    along the z-axis

    Parameters
    ----------
    image : Image
        The input image

    Returns
    -------
    Image
        The projected image

    """
    image_data = image.data.max(dim="z")
    return from_array_like(
        image_data, name="Max Intensity Projection" + image.name, derived_views=[PartialDerivedViewInput(originImage=image)]
    )



import typing
from rekuest_next.structures.default import get_default_structure_registry
import subprocess
import sys
import webbrowser
import os
from rath.scalars import ID


def open_orkestrator_link(link: str):
    webbrowser.open(link)


class IDBearer(typing.Protocol):
    id: ID


def inspect(x: IDBearer) -> None:
    """
    Inspects the given object by opening it in the orkestrator.



    """
    structure_reg = get_default_structure_registry()

    identifier = structure_reg.get_identifier_for_cls(type(x))

    open_orkestrator_link(f"orkestrator://{identifier}/{x.id}")



import contextvars
from fakts_next import Fakts
from koil.composition.base import KoiledModel
from fakts_next.models import Manifest, Requirement
from typing import Any, Callable, Dict, Optional, Protocol, Set, TypeVar
from typing import runtime_checkable
from pydantic import BaseModel

Params = Dict[str, str]


current_service_registry = contextvars.ContextVar(
    "current_service_registry", default=None
)


class Registration(BaseModel):
    name: str
    requirement: Requirement
    builder: Callable[[Fakts, Params], object]
    schema_loader: Callable[[str], object]


@runtime_checkable
class ArkitektService(Protocol):
    def get_service_name(self) -> str:
        """Get the service name. This is used to identify the service in the
        service registry. The service name should be unique across all services.
        """
        ...

    def build_service(self, fakts: Fakts, params: Params) -> Optional[KoiledModel]:
        """Build the service. This is used to build the service and return
        the service instance. The service instance should be a KoiledModel
        that is used to interact with the service.
        """
        ...

    def get_requirements(self) -> list[Requirement]:
        """Get the requirements for the service. This is used to get the
        requirements for the service. The requirements should be a list of
        Requirement objects that are used to identify the service in the
        service registry. The requirements should be unique across all
        """
        ...

    def get_graphql_schema(self) -> Optional[str]:
        """Get the GraphQL schema for the service. This is used to get the
        GraphQL schema for the service. The GraphQL schema should be a
        GraphQL schema object that is used to interact with the service.
        """
        pass

    def get_turms_project(self) -> Optional[Dict[str, Any]]:
        """Get the Turms project for the service. This is used to get the
        Turms project for the service. The Turms project should be a
        Turms project object that is used to interact with the service.
        """
        pass


class BaseArkitektService:
    """Base class for Arkitekt services. This class is used to define the
    interface for Arkitekt services. It is used to define the service name,
    build the service, and get the requirements for the service."""

    def get_service_name(self) -> str:
        raise NotImplementedError("get_service_name not implemented")

    def build_service(self, fakts: Fakts, params: Params) -> Optional[KoiledModel]:
        raise NotImplementedError("build_service not implemented")

    def get_requirements(self) -> list[Requirement]:
        raise NotImplementedError("get_requirements not implemented")

    def get_graphql_schema(self) -> Optional[str]:
        return None

    def get_turms_project(self) -> Optional[Dict[str, Any]]:
        """Get the Turms project for the service. This is used to get the
        Turms project for the service. The Turms project should be a
        Turms project object that is used to interact with the service.
        """
        return None


basic_requirements = []


class ServiceBuilderRegistry:
    def __init__(self):
        self.service_builders: Dict[str, ArkitektService] = {}
        self.additional_requirements: Dict[str, Requirement] = {}

    def register(
        self,
        service: ArkitektService,
    ):
        name = service.get_service_name()

        if name not in self.service_builders:
            self.service_builders[name] = service
        else:
            raise ValueError(f"Service {name} already registered")

    def register_requirement(self, requirement: Requirement):
        if requirement.key in self.additional_requirements:
            raise ValueError(f"Requirement {requirement.key} already registered)")
        self.additional_requirements[requirement.key] = requirement

    def get(self, name: str) -> Optional[ArkitektService]:
        return self.service_builders.get(name)

    def build_service_map(self, fakts: Fakts, params: Params):
        potentially_needed_services = {
            name: service.build_service(fakts, params)
            for name, service in self.service_builders.items()
        }

        return {
            key: value
            for key, value in potentially_needed_services.items()
            if value is not None
        }

    def get_requirements(self):
        requirements = [
            Requirement(
                key="lok",
                service="live.arkitekt.lok",
                description="An instance of ArkitektNext Lok to authenticate the user",
            )
        ]
        taken_requirements: Set[str] = set()

        for service in self.service_builders.values():
            for requirement in service.get_requirements():
                if requirement.key not in taken_requirements:
                    taken_requirements.add(requirement.key)
                    requirements.append(requirement)

        for requirement in self.additional_requirements.values():
            if requirement.key not in taken_requirements:
                taken_requirements.add(requirement.key)
                requirements.append(requirement)

        sorted_requirements = sorted(requirements, key=lambda x: x.key)

        return sorted_requirements


T = TypeVar("T")


def require(
    key: str,
    service: str,
    description: str | None = None,
    service_registry: Optional[ServiceBuilderRegistry] = None,
) -> Requirement:
    """Register a requirement with the service registry

    Parameters
    ----------
    key : str
        The key for the requirement. This should be unique across all
        requirements.

    service : str
        The service that you require. This should be a uinque fakts
        service name. I.e `live.arkitekt.lok` or `live.arkitekt.lok:0.0.1`

    description : str | None
        The description for the requirement. This should be a short
        description of the requirement that gets displayed to the user.

    service_registry : ServiceBuilderRegistry | None
        The service registry to register the requirement with. If
        None, the default service registry will be used.

    Returns
    -------
    Requirement
        The requirement that was registered. This can be used to
        get the requirement later on.

    """
    service_hook_registry = service_registry or get_default_service_registry()

    requirement = Requirement(key=key, service=service, description=description)
    service_hook_registry.register_requirement(requirement)

    return requirement


GLOBAL_SERVICE_REGISTRY = None


def get_default_service_registry() -> "ServiceBuilderRegistry":
    global GLOBAL_SERVICE_REGISTRY
    if GLOBAL_SERVICE_REGISTRY is None:
        GLOBAL_SERVICE_REGISTRY = ServiceBuilderRegistry()  # type: ignore
    return GLOBAL_SERVICE_REGISTRY



from typing import Optional

from fakts_next.fakts import Fakts
from fakts_next.grants.remote import RemoteGrant
from fakts_next.grants.remote.discovery.well_known import WellKnownDiscovery
from fakts_next.grants.remote.demanders.static import StaticDemander
from fakts_next.grants.remote.demanders.device_code import (
    ClientKind,
    DeviceCodeDemander,
    DeviceCodeHook,
    display_in_terminal,
    
)
from fakts_next.grants.remote.claimers.post import ClaimEndpointClaimer
from fakts_next.grants.remote.demanders.redeem import RedeemDemander
from fakts_next.cache.file import FileCache
from fakts_next.models import Manifest


def build_device_code_fakts(
    manifest: Manifest,
    url: Optional[str] = None,
    no_cache: bool = False,
    headless: bool = False,
    device_code_hook: Optional[DeviceCodeHook] = None,
) -> Fakts:
    """ Builds a Fakts instance for device code authentication.
    
    This is used when the user wants to authenticate an application
    using a device code. The user will be prompted to open a browser
    and enter a code to authenticate the application.
    
    
    
    """
    identifier = manifest.identifier
    version = manifest.version
    if url is None:
        raise ValueError("URL must be provided")

    demander = DeviceCodeDemander(
        manifest=manifest,
        open_browser=not headless,
        requested_client_kind=ClientKind.DEVELOPMENT,
        device_code_hook=device_code_hook if device_code_hook else display_in_terminal,
    )

    return Fakts(
        grant=RemoteGrant(
            demander=demander,
            discovery=WellKnownDiscovery(url=url, auto_protocols=["https", "http"]),
            claimer=ClaimEndpointClaimer(),
        ),
        manifest=manifest,
        cache=FileCache(
            cache_file=f".arkitekt_next/cache/{identifier}-{version}_fakts_cache.json",
            hash=manifest.hash() + url,
        ),
    )


def build_redeem_fakts(manifest: Manifest, redeem_token: str, url: str) -> Fakts:
    """ Builds a Fakts instance for redeeming a token.
    
    A redeem token is used to register an application with the
    fakts server, and to claim the configuration for the application.
    
    Instead of using a device code, the user can redeem an application
    without user interaction.
    
    """
    identifier = manifest.identifier
    version = manifest.version

    return Fakts(
        manifest=manifest,
        grant=RemoteGrant(
            demander=RedeemDemander(token=redeem_token, manifest=manifest),
            discovery=WellKnownDiscovery(url=url, auto_protocols=["https", "http"]),
            claimer=ClaimEndpointClaimer(),
        ),
        cache=FileCache(
            cache_file=f".arkitekt_next/cache/{identifier}-{version}_fakts_cache.json",
            hash=manifest.hash() + url,
        ),
    )


def build_token_fakts(
    manifest: Manifest,
    token: str,
    url: str,
) -> Fakts:
    """ Builds a Fakts instance for token-based authentication.
    
    This is used when an appllication was previously authenticated
    and the user has a (claim) token to use for authentication.
    
    E.g. when deploying an application through the kabinet deployer
    
    """
    identifier = manifest.identifier
    version = manifest.version

    return Fakts(
        manifest=manifest,
        grant=RemoteGrant(
            demander=StaticDemander(token=token),  # type: ignore
            discovery=WellKnownDiscovery(url=url, auto_protocols=["https", "http"]),
            claimer=ClaimEndpointClaimer(),
        ),
        cache=FileCache(
            cache_file=f".arkitekt_next/cache/{identifier}-{version}_fakts_cache.json",
            hash=manifest.hash() + url,
        ),
    )



import rich_click as click


import rich_click as click
from .variables import variables
from .implementations import implementations
from .requirements import requirements


@click.group()
@click.pass_context
def inspect(ctx):
    """Inspects your arkitekt_next app

    Inspects various parts of your arkitekt_next app. This is useful for debugging
    and development. It also represents methods that are called by the arkitekt_next
    server when you run your app in production mode.

    """


inspect.add_command(variables, "variables")
inspect.add_command(requirements, "requirements")
inspect.add_command(implementations, "implementations")



from .types import Flavour
import rich_click as click
from click import Context
from rich import get_console
from arkitekt_next.utils import create_arkitekt_next_folder
import yaml


try:
    pass
except ImportError as e:
    raise ImportError("Please install rekuest to use this feature") from e

import os


@click.command()
@click.option("--flavour", "-f", help="The flavour to use", default=None)
@click.option(
    "--description",
    "-d",
    help="The description for this flavour to use",
    default="This is a vanilla flavour",
)
@click.option(
    "--overwrite",
    "-o",
    help="Should we overwrite the existing Dockerfile?",
    is_flag=True,
    default=False,
)
@click.pass_context
def validate(ctx: Context, description: str, overwrite: bool, flavour: str) -> None:
    """Runs the port wizard to generate a dockerfile to be used with port"""

    arkitekt_next_folder = create_arkitekt_next_folder()

    flavours_folder = os.path.join(arkitekt_next_folder, "flavours")

    if not os.path.exists(flavours_folder):
        raise click.ClickException(
            "We could not find the flavours folder. Please run arkitekt_next port init first"
        )

    for dir_name in os.listdir(flavours_folder):
        dir = os.path.join(flavours_folder, dir_name)
        if os.path.isdir(dir):
            if flavour is not None and flavour != dir_name:
                continue

            if os.path.exists(os.path.join(dir, "config.yaml")):
                with open(os.path.join(dir, "config.yaml")) as f:
                    valued = yaml.load(f, Loader=yaml.SafeLoader)
                try:
                    flavour = Flavour(**valued)

                except Exception as e:
                    get_console().print_exception()
                    raise click.ClickException(f"Flavour {dir_name} is invalid") from e

                try:
                    flavour.check_relative_paths(dir)
                except Exception as e:
                    raise click.ClickException(
                        f"Relative Paths in Flavour {dir_name} are invalid. {e}"
                    ) from e

            else:
                raise click.ClickException(
                    f"Flavour {dir_name} is invalid. No config.yaml file found"
                )

        else:
            print(
                f"Found file {dir_name} in flavours folder. Not a valid flavour. Ignoring"
            )

    click.echo("Ice Ice Baby! All flavours are valid")



import contextvars
from functools import wraps
from typing import Callable, Dict, Optional, TypeVar, overload
from .app import App

Params = Dict[str, str]


current_init_hook_registry = contextvars.ContextVar(
    "current_init_hook_registry", default=None
)


InitHook = Callable[[App], None]


class InitHookRegistry:
    """A registry for init hooks. This is used to register init hooks that
    are called when the app is initialized. The init hooks are called in
    the order they are registered

    The purpose of init hooks is to allow specific initialization
    code to be run when the app is initialized. This is useful if you
    plan to add some custom configuration or setup code that needs to be
    run before the app is getting conneted to the server.

    """

    def __init__(self) -> None:
        """Initialize the InitHookRegistry."""
        self.init_hooks: Dict[str, InitHook] = {}
        self.cli_only_hooks: Dict[str, InitHook] = {}

    def register(
        self,
        function: InitHook,
        name: Optional[str] = None,
        only_cli: bool = False,
    ) -> None:
        """Register a function as an init hook. This function will be called


        when the app is initialized. The init hooks are called in the order


        """

        if name is None:
            name = function.__name__

        if only_cli:
            if name not in self.cli_only_hooks:
                self.cli_only_hooks[name] = function
            else:
                raise ValueError(f"CLI Hook {name} already registered")

        if name not in self.init_hooks:
            self.init_hooks[name] = function
        else:
            raise ValueError(f"Init Hook {name} already registered")

    def run_all(self, app: App, is_cli: bool = False) -> None:
        """ Run all registered init hooks."""
        for hook in self.init_hooks.values():
            hook(app)

        if is_cli:
            for hook in self.cli_only_hooks.values():
                hook(app)


T = TypeVar("T", bound=InitHook)


@overload
def init(func: T) -> T: ...


@overload
def init(
    *, only_cli: bool = False, init_hook_registry: InitHookRegistry | None = None
) -> Callable[[T], T]: ...


def init(
    func: T | None = None,
    *,
    only_cli: bool = False,
    init_hook_registry: InitHookRegistry | None = None,
) -> T | Callable[[T], T]:
    """Register a function as an init hook. This function will be called when the app is initialized."""
    init_hook_registry = init_hook_registry or get_default_init_hook_registry()

    if func is not None:
        init_hook_registry.register(func, only_cli=only_cli)
        setattr(func, "__is_init_hook__", True)
        return func

    def decorator(inner: T) -> T:
        @wraps(inner)
        def wrapped(app: App) -> None:
            return inner(app)

        init_hook_registry.register(wrapped, only_cli=only_cli)
        setattr(inner, "__is_init_hook__", True)
        return inner

    return decorator


GLOBAL_INIT_HOOK_REGISTRY = None


def get_default_init_hook_registry()-> InitHookRegistry:
    """Get the default init hook registry. This is used to register init hooks
    that are called when the app is initialized. If no registry is set, a new
    registry is created and returned.
    """
    global GLOBAL_INIT_HOOK_REGISTRY
    if GLOBAL_INIT_HOOK_REGISTRY is None:
        GLOBAL_INIT_HOOK_REGISTRY = InitHookRegistry()  # type: ignore
    return GLOBAL_INIT_HOOK_REGISTRY



from .builders import easy, interactive
from .app.app import App
from fakts_next.helpers import afakt, fakt
from .init_registry import init, InitHookRegistry, get_default_init_hook_registry
from .service_registry import (
    require,
    ServiceBuilderRegistry,
    get_default_service_registry,
)


def missing_install(name: str, error: Exception):
    def real_missing_install(*args, **kwargs):
        raise ImportError(
            f"Missing import: {name}. Please install the missing package. "
        ) from error

    return real_missing_install


try:
    from rekuest_next.register import register
    from rekuest_next.agents.hooks.background import background
    from rekuest_next.agents.hooks.startup import startup
    from rekuest_next.agents.context import context
    from rekuest_next.state.decorator import state
    from rekuest_next.actors.context import abreakpoint, breakpoint
    from rekuest_next.actors.context import progress, aprogress
    from rekuest_next.actors.context import log, alog
    from rekuest_next.structures.model import model
    from rekuest_next.actors.context import apublish, publish
    from rekuest_next.remote import (
        call,
        acall,
        acall_raw,
        iterate,
        aiterate,
        aiterate_raw,
        find,
    )
    from .inspect import inspect
except ImportError as e:
    raise e
    inspect = missing_install("rekuest_next", e)
    publish = missing_install("rekuest_next", e)
    apublish = missing_install("rekuest_next", e)
    structure = missing_install("rekuest_next", e)
    register = missing_install("rekuest_next", e)
    background = missing_install("rekuest_next", e)
    abreakpoint = missing_install("rekuest_next", e)
    breakpoint = missing_install("rekuest_next", e)
    startup = missing_install("rekuest_next", e)
    context = missing_install("rekuest_next", e)
    find = missing_install("rekuest_next", e)
    state = missing_install("rekuest_next", e)
    progress = missing_install("rekuest_next", e)
    aprogress = missing_install("rekuest_next", e)
    log = missing_install("rekuest_next", e)
    alog = missing_install("rekuest_next", e)
    call = missing_install("rekuest_next", e)
    call_raw = missing_install("rekuest_next", e)
    acall = missing_install("rekuest_next", e)
    acall_raw = missing_install("rekuest_next", e)
    find = missing_install("rekuest_next", e)
    afind = missing_install("rekuest_next", e)
    aiterate_raw = missing_install("rekuest_next", e)
    aiterate = missing_install("rekuest_next", e)
    iterate = missing_install("rekuest_next", e)


__all__ = [
    "App",
    "require",
    "easy",
    "interactive",
    "log",
    "alog",
    "afakt",
    "fakt",
    "progress",
    "InitHookRegistry",
    "get_default_init_hook_registry",
    "aprogress",
    "ServiceBuilderRegistry",
    "get_default_service_registry",
    "register",
    "find",
    "breakpoint",
    "abreakpoint",
    "aiterate",
    "inspect",
    "iterate",
    "aiterate_raw",
    "call",
    "acall",
    "acall_raw",
    "model",
    "state",
    "context",
    "background",
    "startup",
    "init",
    "InitHookRegisty",
    "get_current_init_hook_registry",
]



import asyncio
from pydantic import BaseModel
import rich_click as click
from importlib import import_module
from arkitekt_next.app.app import App
from arkitekt_next.cli.commands.run.utils import import_builder
from arkitekt_next.cli.vars import get_console, get_manifest
from arkitekt_next.cli.options import with_builder
import json
import os

from arkitekt_next.constants import DEFAULT_ARKITEKT_URL

try:
    from rekuest_next.definition.registry import get_default_definition_registry
except ImportError:
    get_default_definition_registry = lambda: None
    pass


@click.command("prod")
@click.pass_context
@click.option(
    "--pretty",
    "-p",
    help="Should we just output json?",
    is_flag=True,
    default=False,
)
@click.option(
    "--machine-readable",
    "-mr",
    help="Should we just output json?",
    is_flag=True,
    default=False,
)
def implementations(
    ctx,
    pretty: bool,
    machine_readable: bool,
    builder: str = "arkitekt_next.builders.easy",
    url: str = DEFAULT_ARKITEKT_URL,
):
    """Runs the app in production mode

    \n
    You can specify the builder to use with the --builder flag. By default, the easy builder is used, which is designed to be easy to use and to get started with.

    """

    manifest = get_manifest(ctx)
    console = get_console(ctx)

    entrypoint = manifest.entrypoint
    identifier = manifest.identifier
    entrypoint_file = f"{manifest.entrypoint}.py"
    os.path.realpath(entrypoint_file)

    builder_func = import_builder(builder)

    entrypoint = manifest.entrypoint

    with console.status("Loading entrypoint module..."):
        try:
            import_module(entrypoint)
        except ModuleNotFoundError as e:
            console.print(f"Could not find entrypoint module {entrypoint}")
            raise e

    app: App = builder_func(
        identifier=identifier,
        version="dev",
        logo=manifest.logo,
        url=url,
        headless=True,
    )

    rekuest = app.services.get("rekuest")

    registry = get_default_definition_registry()
    global_list = []
    if registry is None:
        raise Exception("No default registry found")

    to_be_created_templates = tuple(
        x.model_dump() for x in registry.implementations.values()
    )
    global_list.extend(to_be_created_templates)

    console.print(f"Templates to be created: {len(global_list)}")

    if rekuest is None:
        console.print("No rekuest service found in app")
        return

    if machine_readable:
        print(
            "--START_TEMPLATES--"
            + json.dumps(global_list, indent=2)
            + "--END_TEMPLATES--"
        )

    else:
        if pretty:
            console.print(json.dumps(global_list, indent=2))
        else:
            print(json.dumps(global_list))



""" Calling functions in your arkitekt_next app"""

from click import Context
import rich_click as click


from .remote import remote


@click.group()
@click.pass_context
def call(ctx: Context) -> None:
    """Call functions in your arkitekt_next app.
    
    This command allows you to call functions in your app 
    either locally or remotely.
    
    Locally, you can call functions that are defined in your app
    without needing to use rekuest or fakts. (No Server for assignment needed)
    
    Remotely, you can call functions that are defined in your app
    using rekuest or fakts. This goes through a rekuest server to
    remote call the function.
    
    """


call.add_command(remote, "remote")



import rich_click as click
import os


@click.command()
@click.argument("project", default=None, required=False)
@click.option(
    "--config", help="The config to use", type=click.Path(exists=True), default=None
)
def watch(project, config):
    """Watch your projects documents and automatically generate code when they change

    This command will watch all the projects in your config file and automatically
    generate code when the documents change. This is useful for development.

    """
    app_directory = os.getcwd()

    from turms.run import scan_folder_for_single_config, load_projects_from_configpath
    from turms.cli.main import watch_projects

    config = config or scan_folder_for_single_config(app_directory)
    if not config:
        raise click.ClickException(
            f"No config file found. Please run `arkitekt_next gen init` in {app_directory} to create a default config file or specify a config file with the --config flag"
        )

    projects = load_projects_from_configpath(config)
    if project:
        projects = {key: value for key, value in projects.items() if key == project}

    watch_projects(projects, title="ArkitektNext Code Watch")



from rich.console import Group
from rich.tree import Tree
from rich.panel import Panel
from arkitekt_next.app import App
from typing import MutableSet, Tuple, Any, Dict
import os
from .texts import LOGO, WELCOME_MESSAGE

try:
    from rekuest_next.definition.registry import get_default_definition_registry
except ImportError:
    get_default_definition_registry = lambda: None
    pass


def construct_codegen_welcome_panel() -> Panel:
    md = Panel(
        LOGO
        + "[white]"
        + WELCOME_MESSAGE
        + "\n\n"
        + "[bold green]Let's setup your codegen environment",
        title="Welcome to ArkitektNext Codegen",
        title_align="center",
        border_style="green",
        style="green",
    )
    return md


def construct_changes_group(changes: MutableSet[Tuple[Any, str]]) -> Group:
    """Construct a rich panel group for the detected changes

    This panel is displayed if the app has detected changes before
    running the app. This can be caused by a change in the code or
    a change in the environment (e.g. a new package was installed)"""
    panel_header = "Detected changes\n"

    actor_tree = Tree("Changes in", style="white not bold")
    panel_group = Group(panel_header, actor_tree)
    for change, path in changes:
        actor_tree.add(os.path.normpath(path))

    return panel_group


def construct_app_group(app: App) -> Group:
    """Construct a rich panel group for the app

    It displays the registered definitions

    Parameters
    ----------
    app : App
        The app to construct the panel for

    Returns
    -------
    Group
        A rich panel group
    """
    panel_header = f"Running App \n\n{app.fakts.manifest.identifier}:{app.fakts.manifest.version}\n"

    actor_tree = Tree("Registered Definitions", style="white not bold")
    service_tree = Tree("Depends on services", style="white not bold")

    for key, extension in app.services.items():
        service_tree.add(key)

    rekuest = app.services.get("rekuest")
    if rekuest is None:
        return Group(panel_header, service_tree)

    default = get_default_definition_registry()
    if default is not None:
        for key, template in default.implementations.items():
            actor_tree.add(key + "-" + template.definition.name)

    panel_group = Group(panel_header, service_tree, actor_tree)

    return panel_group


def construct_leaking_group(variables: Dict[str, Any]) -> Group:
    """Construct a rich panel group for the leaking variables

    This panel is displayed if the app has leaking variables
    and is therefore considered to be not safe to run
    as an ArkitektNext plugin

    Parameters
    ----------
    variables : Dict[str, Any]
        The leaking variables

    Returns
    -------
    Group
        The rich panel group
    """
    panel_header = (
        "Detected leaking Variables\n\n"
        "[white]Your app is leaking variables. "
        "Leaking variables can cause memory leaks and other issues."
        "Please make sure you are not defining variables in the"
        "global scope (outside of functions). "
        "If you want to define constants please make them all UPPERCASE\n\n"
    )

    actor_tree = Tree("Leaking Variables", style="white not bold")
    panel_group = Group(panel_header, actor_tree)
    for key, value in variables.items():
        actor_tree.add("key: " + key + " value: " + str(value))

    return panel_group


def construct_run_panel(app: App) -> Panel:
    app_group = construct_app_group(app)

    return Panel(
        app_group, style="bold green", border_style="green", title="ArkitektNext Run"
    )
